from __future__ import unicode_literals

import csv
import io
import json
import os
import string

from collections import OrderedDict
from unittest import TestCase

import pandas as pd

from backports.tempfile import TemporaryDirectory
from tempfile import NamedTemporaryFile

from hypothesis import (
    given,
    HealthCheck,
    reproduce_failure,
    settings,
)
from hypothesis.strategies import (
    dictionaries,
    integers,
    just,
    lists,
    text,
    tuples,
)
from mock import patch, Mock

from oasislmf.exposures.manager import OasisExposuresManager
from oasislmf.exposures.pipeline import OasisFilesPipeline
from oasislmf.utils.coverage import (
    BUILDING_COVERAGE_CODE,
    CONTENTS_COVERAGE_CODE,
    OTHER_STRUCTURES_COVERAGE_CODE,
    TIME_COVERAGE_CODE,
)
from oasislmf.utils.exceptions import OasisException
from oasislmf.utils.status import (
    KEYS_STATUS_FAIL,
    KEYS_STATUS_NOMATCH,
    KEYS_STATUS_SUCCESS,
)
from ..models.fakes import fake_model

from tests import (
    canonical_exposure_data,
    keys_data,
    write_input_files,
)


class OasisExposureManagerAddModel(TestCase):
    def test_models_is_empty___model_is_added_to_model_dict(self):
        model = fake_model('supplier', 'model', 'version')

        manager = OasisExposuresManager()
        manager.add_model(model)

        self.assertEqual({model.key: model}, manager.models)

    def test_manager_already_contains_a_model_with_the_given_key___model_is_replaced_in_models_dict(self):
        first = fake_model('supplier', 'model', 'version')
        second = fake_model('supplier', 'model', 'version')

        manager = OasisExposuresManager(oasis_models=[first])
        manager.add_model(second)

        self.assertIs(second, manager.models[second.key])

    def test_manager_already_contains_a_diferent_model___model_is_added_to_dict(self):
        first = fake_model('first', 'model', 'version')
        second = fake_model('second', 'model', 'version')

        manager = OasisExposuresManager(oasis_models=[first])
        manager.add_model(second)

        self.assertEqual({
            first.key: first,
            second.key: second,
        }, manager.models)


class OasisExposureManagerDeleteModels(TestCase):
    def test_models_is_not_in_manager___no_model_is_removed(self):
        manager = OasisExposuresManager([
            fake_model('supplier', 'model', 'version'),
            fake_model('supplier2', 'model2', 'version2'),
        ])
        expected = manager.models

        manager.delete_models([fake_model('supplier3', 'model3', 'version3')])

        self.assertEqual(expected, manager.models)

    def test_models_exist_in_manager___models_are_removed(self):
        models = [
            fake_model('supplier', 'model', 'version'),
            fake_model('supplier2', 'model2', 'version2'),
            fake_model('supplier3', 'model3', 'version3'),
        ]

        manager = OasisExposuresManager(models)
        manager.delete_models(models[1:])

        self.assertEqual({models[0].key: models[0]}, manager.models)


class OasisExposureManagerLoadCanonicalExposuresProfile(TestCase):
    def test_model_and_kwargs_are_not_set___result_is_empty_dict(self):
        profile = OasisExposuresManager().load_canonical_exposures_profile()

        self.assertEqual(None, profile)

    @given(dictionaries(text(), text()))
    def test_model_is_set_with_profile_json___models_profile_is_set_to_expected_json(self, expected):
        model = fake_model(resources={'canonical_exposures_profile_json': json.dumps(expected)})

        profile = OasisExposuresManager().load_canonical_exposures_profile(oasis_model=model)

        self.assertEqual(expected, profile)
        self.assertEqual(expected, model.resources['canonical_exposures_profile'])

    @given(dictionaries(text(), text()), dictionaries(text(), text()))
    def test_model_is_set_with_profile_json_and_profile_json_is_passed_through_kwargs___kwargs_profile_is_used(
        self,
        model_profile,
        kwargs_profile
    ):
        model = fake_model(resources={'canonical_exposures_profile_json': json.dumps(model_profile)})

        profile = OasisExposuresManager().load_canonical_exposures_profile(oasis_model=model, canonical_exposures_profile_json=json.dumps(kwargs_profile))

        self.assertEqual(kwargs_profile, profile)
        self.assertEqual(kwargs_profile, model.resources['canonical_exposures_profile'])

    @given(dictionaries(text(), text()))
    def test_model_is_set_with_profile_json_path___models_profile_is_set_to_expected_json(self, expected):
        with NamedTemporaryFile('w') as f:
            json.dump(expected, f)
            f.flush()

            model = fake_model(resources={'canonical_exposures_profile_json_path': f.name})

            profile = OasisExposuresManager().load_canonical_exposures_profile(oasis_model=model)

            self.assertEqual(expected, profile)
            self.assertEqual(expected, model.resources['canonical_exposures_profile'])

    @given(dictionaries(text(), text()), dictionaries(text(), text()))
    def test_model_is_set_with_profile_json_path_and_profile_json_path_is_passed_through_kwargs___kwargs_profile_is_used(
        self,
        model_profile,
        kwargs_profile
    ):
        with NamedTemporaryFile('w') as model_file, NamedTemporaryFile('w') as kwargs_file:
            json.dump(model_profile, model_file)
            model_file.flush()
            json.dump(kwargs_profile, kwargs_file)
            kwargs_file.flush()

            model = fake_model(resources={'canonical_exposures_profile_json_path': model_file.name})

            profile = OasisExposuresManager().load_canonical_exposures_profile(oasis_model=model, canonical_exposures_profile_json_path=kwargs_file.name)

            self.assertEqual(kwargs_profile, profile)
            self.assertEqual(kwargs_profile, model.resources['canonical_exposures_profile'])


class OasisExposureManagerGetKeys(TestCase):
    def create_model(
        self,
        lookup='lookup',
        keys_file_path='key_file_path',
        keys_errors_file_path='keys_errors_file_path',
        model_exposures_file_path='model_exposures_file_path'
    ):
        model = fake_model(resources={'lookup': lookup})
        model.resources['oasis_files_pipeline'].keys_file_path = keys_file_path
        model.resources['oasis_files_pipeline'].keys_errors_file_path = keys_errors_file_path
        model.resources['oasis_files_pipeline'].model_exposures_file_path = model_exposures_file_path
        return model

    @given(
        lookup=text(min_size=1, alphabet=string.ascii_letters),
        keys_file_path=text(min_size=1, alphabet=string.ascii_letters),
        keys_errors_file_path=text(min_size=1, alphabet=string.ascii_letters),
        exposures_file_path=text(min_size=1, alphabet=string.ascii_letters)
    )
    def test_model_is_supplied_kwargs_are_not___lookup_keys_files_and_exposures_file_from_model_are_used(
        self,
        lookup,
        keys_file_path,
        keys_errors_file_path,
        exposures_file_path
    ):
        model = self.create_model(lookup=lookup, keys_file_path=keys_file_path, keys_errors_file_path=keys_errors_file_path, model_exposures_file_path=exposures_file_path)

        with patch('oasislmf.exposures.manager.OasisLookupFactory.save_results', Mock(return_value=(keys_file_path, 1, keys_errors_file_path, 1))) as oklf_mock:
            res_keys_file_path, res_keys_errors_file_path = OasisExposuresManager().get_keys(oasis_model=model)

            oklf_mock.assert_called_once_with(
                lookup,
                keys_file_path,
                errors_fp=keys_errors_file_path,
                model_exposures_fp=exposures_file_path
            )
            self.assertEqual(model.resources['oasis_files_pipeline'].keys_file_path, keys_file_path)
            self.assertEqual(res_keys_file_path, keys_file_path)
            self.assertEqual(model.resources['oasis_files_pipeline'].keys_errors_file_path, keys_errors_file_path)
            self.assertEqual(res_keys_errors_file_path, keys_errors_file_path)

    @given(
        model_lookup=text(min_size=1, alphabet=string.ascii_letters), 
        model_keys_fp=text(min_size=1, alphabet=string.ascii_letters),
        model_keys_errors_fp=text(min_size=1, alphabet=string.ascii_letters),
        model_exposures_fp=text(min_size=1, alphabet=string.ascii_letters),
        lookup=text(min_size=1, alphabet=string.ascii_letters),
        keys_fp=text(min_size=1, alphabet=string.ascii_letters),
        keys_errors_fp=text(min_size=1, alphabet=string.ascii_letters),
        exposures_fp=text(min_size=1, alphabet=string.ascii_letters)
    )
    def test_model_and_kwargs_are_supplied___lookup_keys_files_and_exposures_file_from_kwargs_are_used(
        self,
        model_lookup,
        model_keys_fp,
        model_keys_errors_fp,
        model_exposures_fp,
        lookup,
        keys_fp,
        keys_errors_fp,
        exposures_fp
    ):
        model = self.create_model(lookup=model_lookup, keys_file_path=model_keys_fp, keys_errors_file_path=model_keys_errors_fp, model_exposures_file_path=model_exposures_fp)

        with patch('oasislmf.exposures.manager.OasisLookupFactory.save_results', Mock(return_value=(keys_fp, 1, keys_errors_fp, 1))) as oklf_mock:
            res_keys_file_path, res_keys_errors_file_path = OasisExposuresManager().get_keys(
                oasis_model=model,
                lookup=lookup,
                model_exposures_file_path=exposures_fp,
                keys_file_path=keys_fp,
                keys_errors_file_path=keys_errors_fp
            )

            oklf_mock.assert_called_once_with(
                lookup,
                keys_fp,
                errors_fp=keys_errors_fp,
                model_exposures_fp=exposures_fp
            )
            self.assertEqual(model.resources['oasis_files_pipeline'].keys_file_path, keys_fp)
            self.assertEqual(res_keys_file_path, keys_fp)
            self.assertEqual(model.resources['oasis_files_pipeline'].keys_errors_file_path, keys_errors_fp)
            self.assertEqual(res_keys_errors_file_path, keys_errors_fp)


class OasisExposureManagerLoadMasterDataframe(TestCase):
    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        profile_element_name=text(alphabet=string.ascii_letters, min_size=1),
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_row_in_keys_data_is_missing_from_exposure_data___oasis_exception_is_raised(
        self,
        profile_element_name,
        keys,
        exposures
    ):
        matching_exposures = [e for e in exposures if e[0] in map(lambda k: k['id'], keys)] 
        exposures.pop(exposures.index(matching_exposures[0]))
        profile = {
            profile_element_name: {'ProfileElementName': profile_element_name, 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name, profile_element_name=profile_element_name)

            with self.assertRaises(OasisException):
                OasisExposuresManager().load_master_data_frame(exposures_file.name, keys_file.name, profile)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        profile_element_name=text(alphabet=string.ascii_letters, min_size=1),
        keys=keys_data(from_coverage_type_ids=just(CONTENTS_COVERAGE_CODE), from_statuses=just(KEYS_STATUS_SUCCESS), size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_canonical_profile_coverage_types_dont_match_model_defined_coverage_types___oasis_exception_is_raised(
        self,
        profile_element_name,
        keys,
        exposures
    ):
        matching_exposures = [e for e in exposures if e[0] in map(lambda k: k['id'], keys)] 
        exposures.pop(exposures.index(matching_exposures[0]))
        profile = {
            profile_element_name: {'ProfileElementName': profile_element_name, 'FieldName': 'TIV', 'CoverageTypeID': BUILDING_COVERAGE_CODE}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name, profile_element_name=profile_element_name)

            with self.assertRaises(OasisException):
                OasisExposuresManager().load_master_data_frame(exposures_file.name, keys_file.name, profile)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        profile_element_name=text(alphabet=string.ascii_letters, min_size=1),
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), size=10),
        exposures=canonical_exposure_data(num_rows=10, min_value=1)
    )
    def test_each_row_has_a_single_row_per_element_with_each_row_having_a_positive_value_for_the_profile_element___each_row_is_present(
        self,
        profile_element_name,
        keys,
        exposures
    ):
        profile = {
            profile_element_name: {'ProfileElementName': profile_element_name, 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        expected = []
        keys_values_tuples = map(lambda li: tuple(filter(lambda v: type(v) == int, li)), [k.values() for k in keys])
        for i, zipped_data in enumerate(zip(keys_values_tuples, exposures)):
            expected.append((
                i + 1,
                zipped_data[0],
                zipped_data[1][1],
            ))

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name, profile_element_name=profile_element_name)

            result = OasisExposuresManager().load_master_data_frame(
                exposures_file.name,
                keys_file.name,
                profile,
            )
        self.assertEqual(len(expected), len(result))

        for i in range(len(result)):
            row = {k:(int(v) if k != 'tiv' else v) for k, v in result.iloc[i].to_dict().items()}
            self.assertEqual(i + 1, row['item_id'])
            self.assertEqual(i + 1, row['coverage_id'])
            self.assertEqual(exposures[i][1], row['tiv'])
            self.assertEqual(keys[i]['area_peril_id'], row['areaperil_id'])
            self.assertEqual(keys[i]['vulnerability_id'], row['vulnerability_id'])
            self.assertEqual(i + 1, row['group_id'])
            self.assertEqual(1, row['summary_id'])
            self.assertEqual(1, row['summaryset_id'])

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        profile_element_name=text(alphabet=string.ascii_letters, min_size=1),
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), size=10),
        exposures=canonical_exposure_data(num_rows=10, min_value=1)
    )
    def test_each_row_has_a_single_row_per_element_with_each_row_having_any_value_for_the_profile_element___rows_with_profile_elements_gt_0_are_present(
        self,
        profile_element_name,
        keys,
        exposures
    ):
        profile = {
            profile_element_name: {'ProfileElementName': profile_element_name, 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        expected = []
        keys_values_tuples = map(lambda li: tuple(filter(lambda v: type(v) == int, li)), [k.values() for k in keys])
        row_id = 0
        for zipped_keys, zipped_exposure in zip(keys_values_tuples, exposures):
            if zipped_exposure[1] > 0:
                row_id += 1
                expected.append((
                    row_id,
                    zipped_keys,
                    zipped_exposure[1],
                ))

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name, profile_element_name=profile_element_name)

            result = OasisExposuresManager().load_master_data_frame(
                exposures_file.name,
                keys_file.name,
                profile,
            )

        self.assertEqual(len(expected), len(result))
        for i in range(len(result)):
            row = {k:(int(v) if k != 'tiv' else v) for k, v in result.iloc[i].to_dict().items()}
            self.assertEqual(i + 1, row['item_id'])
            self.assertEqual(i + 1, row['coverage_id'])
            self.assertEqual(exposures[i][1], row['tiv'])
            self.assertEqual(keys[i]['area_peril_id'], row['areaperil_id'])
            self.assertEqual(keys[i]['vulnerability_id'], row['vulnerability_id'])
            self.assertEqual(i + 1, row['group_id'])
            self.assertEqual(1, row['summary_id'])
            self.assertEqual(1, row['summaryset_id'])


class FileGenerationTestCase(TestCase):
    def setUp(self):
        self.items_filename = 'items.csv'
        self.coverages_filename = 'coverages.csv'
        self.gulsummaryxref_filename = 'gulsummaryxref.csv'

    def check_items_file(self, keys, out_dir):
        expected = [
            {
                'item_id': i + 1,
                'coverage_id': i + 1,
                'areaperil_id': key['area_peril_id'],
                'vulnerability_id': key['vulnerability_id'],
                'group_id': i + 1,
            } for i, key in enumerate(keys)
        ]

        with io.open(os.path.join(out_dir, self.items_filename), 'r', encoding='utf-8') as f:
            result = list(pd.read_csv(f).T.to_dict().values())
        
        self.assertEqual(expected, result)

    def check_coverages_file(self, exposures, out_dir):
        expected = [
            {
                'coverage_id': item_id + 1,
                'tiv': item[1],
            } for item_id, item in enumerate(exposures)
        ]

        with io.open(os.path.join(out_dir, self.coverages_filename), 'r', encoding='utf-8') as f:
            result = list(pd.read_csv(f).T.to_dict().values())
        
        self.assertEqual(expected, result)

    def check_gul_file(self, exposures, out_dir):
        expected = [
            {
                'coverage_id': item_id + 1,
                'summary_id': 1,
                'summaryset_id': 1,
            } for item_id in range(len(exposures))
        ]

        with io.open(os.path.join(out_dir, self.gulsummaryxref_filename), 'r', encoding='utf-8') as f:
            result = list(pd.read_csv(f).T.to_dict().values())
        
        self.assertEqual(expected, result)


class OasisExposuresManagerGenerateItemsFile(FileGenerationTestCase):

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_model___model_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model(resources={'canonical_exposures_profile': profile})
            model.resources['oasis_files_pipeline'].keys_file_path = keys_file.name
            model.resources['oasis_files_pipeline'].canonical_exposures_file_path = exposures_file.name
            model.resources['oasis_files_pipeline'].items_file_path = os.path.join(out_dir, self.items_filename)

            OasisExposuresManager().generate_items_file(oasis_model=model)

            self.check_items_file(keys, out_dir)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_kwargs___kwarg_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model()

            OasisExposuresManager().generate_items_file(
                oasis_model=model,
                canonical_exposures_profile=profile,
                keys_file_path=keys_file.name,
                canonical_exposures_file_path=exposures_file.name,
                items_file_path=os.path.join(out_dir, self.items_filename)
            )

            self.check_items_file(keys, out_dir)


class OasisExposuresManagerGenerateCoveragesFile(FileGenerationTestCase):

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_model___model_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model(resources={'canonical_exposures_profile': profile})
            model.resources['oasis_files_pipeline'].keys_file_path = keys_file.name
            model.resources['oasis_files_pipeline'].canonical_exposures_file_path = exposures_file.name
            model.resources['oasis_files_pipeline'].coverages_file_path = os.path.join(out_dir, self.coverages_filename)

            OasisExposuresManager().generate_coverages_file(oasis_model=model)

            self.check_coverages_file(exposures, out_dir)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_kwargs___kwarg_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model()

            OasisExposuresManager().generate_coverages_file(
                oasis_model=model,
                canonical_exposures_profile=profile,
                keys_file_path=keys_file.name,
                canonical_exposures_file_path=exposures_file.name,
                coverages_file_path=os.path.join(out_dir, self.coverages_filename)
            )

            self.check_coverages_file(exposures, out_dir)


class OasisExposuresManagerGenerateGulsummaryxrefFile(FileGenerationTestCase):

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_model___model_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model(resources={'canonical_exposures_profile': profile})
            model.resources['oasis_files_pipeline'].keys_file_path = keys_file.name
            model.resources['oasis_files_pipeline'].canonical_exposures_file_path = exposures_file.name
            model.resources['oasis_files_pipeline'].gulsummaryxref_file_path = os.path.join(out_dir, self.gulsummaryxref_filename)

            OasisExposuresManager().generate_gulsummaryxref_file(oasis_model=model)

            self.check_gul_file(exposures, out_dir)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_kwargs___kwarg_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model()
            OasisExposuresManager().generate_gulsummaryxref_file(
                oasis_model=model,
                canonical_exposures_profile=profile,
                keys_file_path=keys_file.name,
                canonical_exposures_file_path=exposures_file.name,
                gulsummaryxref_file_path=os.path.join(out_dir, self.gulsummaryxref_filename)
            )

            self.check_gul_file(exposures, out_dir)


class OasisExposuresManagerGenerateOasisFiles(FileGenerationTestCase):
    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_model___model_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model(resources={'canonical_exposures_profile': profile})
            model.resources['oasis_files_pipeline'].keys_file_path = keys_file.name
            model.resources['oasis_files_pipeline'].canonical_exposures_file_path = exposures_file.name
            model.resources['oasis_files_pipeline'].items_file_path = os.path.join(out_dir, self.items_filename)
            model.resources['oasis_files_pipeline'].coverages_file_path = os.path.join(out_dir, self.coverages_filename)
            model.resources['oasis_files_pipeline'].gulsummaryxref_file_path = os.path.join(out_dir, self.gulsummaryxref_filename)

            OasisExposuresManager().generate_oasis_files(oasis_model=model)

            self.check_items_file(keys, out_dir)
            self.check_coverages_file(exposures, out_dir)
            self.check_gul_file(exposures, out_dir)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        keys=keys_data(from_statuses=just(KEYS_STATUS_SUCCESS), min_size=10),
        exposures=canonical_exposure_data(10, min_value=1)
    )
    def test_paths_are_stored_in_the_kwargs___kwarg_paths_are_used(self, keys, exposures):
        profile = {
            'profile_element': {'ProfileElementName': 'profile_element', 'FieldName': 'TIV', 'CoverageTypeID': 1}
        }

        with NamedTemporaryFile('w') as keys_file, NamedTemporaryFile('w') as exposures_file, TemporaryDirectory() as out_dir:
            write_input_files(keys, keys_file.name, exposures, exposures_file.name)

            model = fake_model()

            OasisExposuresManager().generate_oasis_files(
                oasis_model=model,
                canonical_exposures_profile=profile,
                keys_file_path=keys_file.name,
                canonical_exposures_file_path=exposures_file.name,
                items_file_path=os.path.join(out_dir, self.items_filename),
                coverages_file_path=os.path.join(out_dir, self.coverages_filename),
                gulsummaryxref_file_path=os.path.join(out_dir, self.gulsummaryxref_filename)
            )

            self.check_items_file(keys, out_dir)
            self.check_coverages_file(exposures, out_dir)
            self.check_gul_file(exposures, out_dir)


class OasisExposuresTransformSourceToCanonical(TestCase):
    @given(
        source_exposures_file_path=text(),
        source_exposures_validation_file_path=text(),
        source_to_canonical_exposures_transformation_file_path=text(),
        canonical_exposures_file_path=text()
    )
    def test_model_is_not_set___parameters_are_taken_from_kwargs(
            self,
            source_exposures_file_path,
            source_exposures_validation_file_path,
            source_to_canonical_exposures_transformation_file_path,
            canonical_exposures_file_path
    ):
        trans_call_mock = Mock()
        with patch('oasislmf.exposures.manager.Translator', Mock(return_value=trans_call_mock)) as trans_mock:
            OasisExposuresManager().transform_source_to_canonical(
                source_exposures_file_path=source_exposures_file_path,
                source_exposures_validation_file_path=source_exposures_validation_file_path,
                source_to_canonical_exposures_transformation_file_path=source_to_canonical_exposures_transformation_file_path,
                canonical_exposures_file_path=canonical_exposures_file_path
            )

            trans_mock.assert_called_once_with(
                os.path.abspath(source_exposures_file_path),
                os.path.abspath(canonical_exposures_file_path),
                os.path.abspath(source_exposures_validation_file_path),
                os.path.abspath(source_to_canonical_exposures_transformation_file_path),
                append_row_nums=True,
            )
            trans_call_mock.assert_called_once_with()

    @given(
        source_exposures_file_path=text(),
        source_exposures_validation_file_path=text(),
        source_to_canonical_exposures_transformation_file_path=text(),
        canonical_exposures_file_path=text()
    )
    def test_model_is_set___parameters_are_taken_from_model(
            self,
            source_exposures_file_path,
            source_exposures_validation_file_path,
            source_to_canonical_exposures_transformation_file_path,
            canonical_exposures_file_path):

        model = fake_model(resources={
            'source_exposures_file_path': source_exposures_file_path,
            'source_exposures_validation_file_path': source_exposures_validation_file_path,
            'source_to_canonical_exposures_transformation_file_path': source_to_canonical_exposures_transformation_file_path,
        })
        model.resources['oasis_files_pipeline'].canonical_exposures_path = canonical_exposures_file_path

        trans_call_mock = Mock()
        with patch('oasislmf.exposures.manager.Translator', Mock(return_value=trans_call_mock)) as trans_mock:
            OasisExposuresManager().transform_source_to_canonical(
                source_exposures_file_path=source_exposures_file_path,
                source_exposures_validation_file_path=source_exposures_validation_file_path,
                source_to_canonical_exposures_transformation_file_path=source_to_canonical_exposures_transformation_file_path,
                canonical_exposures_file_path=canonical_exposures_file_path
            )

            trans_mock.assert_called_once_with(
                os.path.abspath(source_exposures_file_path),
                os.path.abspath(canonical_exposures_file_path),
                os.path.abspath(source_exposures_validation_file_path),
                os.path.abspath(source_to_canonical_exposures_transformation_file_path),
                append_row_nums=True,
            )
            trans_call_mock.assert_called_once_with()


class OasisExposuresTransformCanonicalToModel(TestCase):
    @given(
        canonical_exposures_file_path=text(),
        canonical_exposures_validation_file_path=text(),
        canonical_to_model_exposures_transformation_file_path=text(),
        model_exposures_file_path=text()
    )
    def test_model_is_not_set___parameters_are_taken_from_kwargs(
            self,
            canonical_exposures_file_path,
            canonical_exposures_validation_file_path,
            canonical_to_model_exposures_transformation_file_path,
            model_exposures_file_path):

        trans_call_mock = Mock()
        with patch('oasislmf.exposures.manager.Translator', Mock(return_value=trans_call_mock)) as trans_mock:
            OasisExposuresManager().transform_canonical_to_model(
                canonical_exposures_file_path=canonical_exposures_file_path,
                canonical_exposures_validation_file_path=canonical_exposures_validation_file_path,
                canonical_to_model_exposures_transformation_file_path=canonical_to_model_exposures_transformation_file_path,
                model_exposures_file_path=model_exposures_file_path,
            )

            trans_mock.assert_called_once_with(
                os.path.abspath(canonical_exposures_file_path),
                os.path.abspath(model_exposures_file_path),
                os.path.abspath(canonical_exposures_validation_file_path),
                os.path.abspath(canonical_to_model_exposures_transformation_file_path),
                append_row_nums=False,
            )
            trans_call_mock.assert_called_once_with()

    @given(
        canonical_exposures_file_path=text(),
        canonical_exposures_validation_file_path=text(),
        canonical_to_model_exposures_transformation_file_path=text(),
        model_exposures_file_path=text()
    )
    def test_model_is_set___parameters_are_taken_from_model(
            self,
            canonical_exposures_file_path,
            canonical_exposures_validation_file_path,
            canonical_to_model_exposures_transformation_file_path,
            model_exposures_file_path):

        model = fake_model(resources={
            'canonical_exposures_validation_file_path': canonical_exposures_validation_file_path,
            'canonical_to_model_exposures_transformation_file_path': canonical_to_model_exposures_transformation_file_path,
        })
        model.resources['oasis_files_pipeline'].canonical_exposures_path = canonical_exposures_file_path
        model.resources['oasis_files_pipeline'].model_exposures_file_path = model_exposures_file_path

        trans_call_mock = Mock()
        with patch('oasislmf.exposures.manager.Translator', Mock(return_value=trans_call_mock)) as trans_mock:
            OasisExposuresManager().transform_canonical_to_model(
                canonical_exposures_file_path=canonical_exposures_file_path,
                canonical_exposures_validation_file_path=canonical_exposures_validation_file_path,
                canonical_to_model_exposures_transformation_file_path=canonical_to_model_exposures_transformation_file_path,
                model_exposures_file_path=model_exposures_file_path,
            )

            trans_mock.assert_called_once_with(
                os.path.abspath(canonical_exposures_file_path),
                os.path.abspath(model_exposures_file_path),
                os.path.abspath(canonical_exposures_validation_file_path),
                os.path.abspath(canonical_to_model_exposures_transformation_file_path),
                append_row_nums=False,
            )
            trans_call_mock.assert_called_once_with()


class OasisExposureManagerCreate(TestCase):
    @given(supplier=text(), model_id=text(), version=text())
    def test_supplier_model_and_version_are_supplied___correct_key_is_created(self, supplier, model_id, version):
        model = fake_model(supplier=supplier, model=model_id, version=version)

        self.assertEqual('{}/{}/{}'.format(supplier, model_id, version), model.key)

    def test_oasis_file_path_is_given___path_is_stored_as_absolute_path(self):
        model = fake_model(resources={'oasis_files_path': 'some_path'})

        result = model.resources['oasis_files_path']
        expected = os.path.abspath('some_path')

        self.assertEqual(expected, result)

    def test_oasis_file_path_is_not_given___path_is_abs_path_of_default(self):
        model = fake_model()

        result = model.resources['oasis_files_path']
        expected = os.path.abspath(os.path.join('Files', model.key.replace('/', '-')))

        self.assertEqual(expected, result)

    def test_file_pipeline_is_not_supplied___default_pipeline_is_set(self):
        model = fake_model()

        pipeline = model.resources['oasis_files_pipeline']

        self.assertIsInstance(pipeline, OasisFilesPipeline)
        self.assertEqual(pipeline.model_key, model.key)

    def test_file_pipeline_is_supplied___pipeline_is_unchanged(self):
        pipeline = OasisFilesPipeline()

        model = fake_model(resources={'oasis_files_pipeline': pipeline})

        self.assertIs(pipeline, model.resources['oasis_files_pipeline'])

    def test_pipeline_is_not_a_pipeline_instance___oasis_exception_is_raised(self):
        class FakePipeline(object):
            pass

        pipeline = FakePipeline()

        with self.assertRaises(OasisException):
            fake_model(resources={'oasis_files_pipeline': pipeline})

    def test_canonical_exposures_profile_not_set___canonical_exposures_profile_in_none(self):
        model = fake_model()

        profile = model.resources['canonical_exposures_profile']

        self.assertEqual(None, profile)

    @given(expected=dictionaries(text(), text()))
    def test_canonical_exposures_profile_json_set___canonical_exposures_profile_matches_json(self, expected):
        model = fake_model(resources={'canonical_exposures_profile_json': json.dumps(expected)})

        profile = model.resources['canonical_exposures_profile']

        self.assertEqual(expected, profile)

    @given(expected=dictionaries(text(), text()))
    def test_canonical_exposures_profile_path_set___canonical_exposures_profile_matches_json(self, expected):
        with NamedTemporaryFile('w') as f:
            json.dump(expected, f)
            f.flush()

            model = fake_model(resources={'canonical_exposures_profile_json_path': f.name})

            profile = model.resources['canonical_exposures_profile']

            self.assertEqual(expected, profile)

    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    @given(expected=dictionaries(text(), text()), new=dictionaries(text(), text()))
    def test_canonical_exposures_profile_set___profile_is_not_updated(self, expected, new):
        model = fake_model(resources={
            'canonical_exposures_profile': expected,
            'canonical_exposures_profile_json': json.dumps(new),
        })

        profile = model.resources['canonical_exposures_profile']

        self.assertEqual(expected, profile)



from __future__ import unicode_literals

__all__ = [
    'coverage_type_ids',
    'keys_data',
    'keys_status_flags',
    'peril_ids',
    'oasis_keys_file_to_record_metadict',
    'oasis_keys_error_file_to_record_metadict'
]

import string

from hypothesis import given
from hypothesis.strategies import (
    booleans,
    fixed_dictionaries,
    integers,
    just,
    lists,
    sampled_from,
    text,
    tuples,
)

from .coverage import (
    BUILDING_COVERAGE_CODE,
    CONTENTS_COVERAGE_CODE,
    OTHER_STRUCTURES_COVERAGE_CODE,
    TIME_COVERAGE_CODE,
)
from .peril import (
    PERIL_ID_FLOOD,
    PERIL_ID_QUAKE,
    PERIL_ID_SURGE,
    PERIL_ID_WIND,
)
from .status import (
    KEYS_STATUS_FAIL,
    KEYS_STATUS_NOMATCH,
    KEYS_STATUS_SUCCESS,
)

coverage_type_ids = [BUILDING_COVERAGE_CODE, CONTENTS_COVERAGE_CODE, OTHER_STRUCTURES_COVERAGE_CODE, TIME_COVERAGE_CODE]

peril_ids = [PERIL_ID_FLOOD, PERIL_ID_QUAKE, PERIL_ID_QUAKE, PERIL_ID_WIND]

keys_status_flags = [KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH, KEYS_STATUS_SUCCESS]


def keys_data(peril_id=PERIL_ID_WIND, coverage_type_id=BUILDING_COVERAGE_CODE, status=None, min_size=1, max_size=10):

    def keys_records(
        from_ids=integers(),
        from_peril_ids=sampled_from(peril_ids),
        from_coverage_type_ids=sampled_from(coverage_type_ids),
        from_area_peril_ids=integers(),
        from_vulnerability_ids=integers(),
        from_statuses=sampled_from(keys_status_flags),
        from_messages=text(min_size=1, max_size=100, alphabet=string.ascii_letters),
        min_size=1,
        max_size=10
    ):
        return lists(fixed_dictionaries({
            'id': from_ids,
            'peril_id': from_peril_ids,
            'coverage': from_coverage_type_ids,
            'area_peril_id': from_area_peril_ids,
            'vulnerability_id': from_vulnerability_ids,
            'status': from_statuses,
            'message': from_messages
        }), min_size=min_size, max_size=max_size)

    if status and status in keys_status_flags and peril_id and coverage_type_id:
        return keys_records(from_statuses=just(status), from_peril_ids=just(peril_id), from_coverage_type_ids=just(coverage_type_id), min_size=min_size, max_size=max_size)

    elif status and status in keys_status_flags and peril_id and not coverage_type_id:
        return keys_records(from_statuses=just(status), from_peril_ids=just(peril_id), from_coverage_type_ids=sampled_from(coverage_type_ids), min_size=min_size, max_size=max_size)

    elif status and status in keys_status_flags and not peril_id and coverage_type_id:
        return keys_records(from_statuses=just(status), from_peril_ids=sampled_from(peril_ids), from_coverage_type_ids=coverage_type_id, min_size=min_size, max_size=max_size)

    elif status and status in keys_status_flags and not (peril_id or coverage_type_id):
        return keys_records(from_statuses=sampled_from([KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH]), from_peril_ids=sampled_from(peril_ids), from_coverage_type_ids=sampled_from(coverage_type_ids), min_size=min_size, max_size=max_size)

    elif status and status not in keys_status_flags and peril_id and coverage_type_id:
        return keys_records(from_statuses=sampled_from([KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH]), from_peril_ids=just(peril_id), from_coverage_type_ids=just(coverage_type_id), min_size=min_size, max_size=max_size)

    elif status and status not in keys_status_flags and peril_id and not coverage_type_id:
        return keys_records(from_statuses=sampled_from([KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH]), from_peril_ids=just(peril_id), from_coverage_type_ids=sampled_from(coverage_type_ids), min_size=min_size, max_size=max_size)

    elif status and status not in keys_status_flags and not peril_id and coverage_type_id:
        return keys_records(from_statuses=sampled_from([KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH]), from_peril_ids=sampled_from(peril_ids), from_coverage_type_ids=just(coverage_type_id), min_size=min_size, max_size=max_size)

    elif status and status not in keys_status_flags and not (peril_id or coverage_type_id):
        return keys_records(from_statuses=sampled_from([KEYS_STATUS_FAIL, KEYS_STATUS_NOMATCH]), from_peril_ids=sampled_from(peril_ids), from_coverage_type_ids=sampled_from(coverage_type_ids), min_size=min_size, max_size=max_size)

    elif not status and peril_id and coverage_type_id:
        return keys_records(from_peril_ids=just(peril_id), from_coverage_type_ids=just(coverage_type_id) min_size=min_size, max_size=max_size)

    elif not status and peril_id and not coverage_type_id:
        return keys_records(from_peril_ids=just(peril_id), from_coverage_type_ids=sampled_from(coverage_type_ids) min_size=min_size, max_size=max_size)

    elif not status and not peril_id and coverage_type_id:
        return keys_records(from_peril_ids=sampled_from(peril_ids), from_coverage_type_ids=just(coverage_type_id) min_size=min_size, max_size=max_size)

    elif not (status or peril_id or coverage_type_id):
        return keys_records(min_size=min_size, max_size=max_size)



from __future__ import unicode_literals

import json
import uuid
from unittest import TestCase

import os
import io

from backports.tempfile import TemporaryDirectory
from hypothesis import given
from hypothesis.strategies import sampled_from
from pathlib2 import Path

from oasislmf.model_execution.conf import create_analysis_settings_json
from oasislmf.model_execution.files import GENERAL_SETTINGS_FILE, MODEL_SETTINGS_FILE, GUL_SUMMARIES_FILE, \
    IL_SUMMARIES_FILE
from oasislmf.utils.exceptions import OasisException


class CreateAnalysisSettingsJson(TestCase):
    def create_fake_directory(self, d):
        Path(os.path.join(d, GENERAL_SETTINGS_FILE)).touch()
        Path(os.path.join(d, MODEL_SETTINGS_FILE)).touch()
        Path(os.path.join(d, GUL_SUMMARIES_FILE)).touch()
        Path(os.path.join(d, IL_SUMMARIES_FILE)).touch()

    def test_directory_does_not_exist___oasis_exception_is_raised(self):
        with self.assertRaises(OasisException):
            create_analysis_settings_json('/tmp/non_existing_dir_{}'.format(uuid.uuid4().hex))

    @given(sampled_from([GENERAL_SETTINGS_FILE, MODEL_SETTINGS_FILE, GUL_SUMMARIES_FILE, IL_SUMMARIES_FILE]))
    def test_input_file_is_missing___error_is_raised(self, to_delete):
        with TemporaryDirectory() as d:
            self.create_fake_directory(d)

            os.remove(os.path.join(d, to_delete))

            with self.assertRaises(OasisException):
                create_analysis_settings_json(d)

    def test_general_settings_file_contains_properties___properties_are_loaded_into_general_settings(self):
        with TemporaryDirectory() as d:
            self.create_fake_directory(d)

            with io.open(os.path.join(d, GENERAL_SETTINGS_FILE), 'w', encoding='utf-8') as f:
                f.writelines([
                    'first,1,int\n',
                    'second,foo,str\n',
                    'third,2.2,float\n',
                ])

            data = json.loads(create_analysis_settings_json(d))

            self.assertEqual(data['first'], 1)
            self.assertEqual(data['second'], 'foo')
            self.assertEqual(data['third'], 2.2)

    def test_model_settings_file_contains_properties___properties_are_loaded_into_model_settings(self):
        with TemporaryDirectory() as d:
            self.create_fake_directory(d)

            with io.open(os.path.join(d, MODEL_SETTINGS_FILE), 'w', encoding='utf-8') as f:
                f.writelines([
                    'first,3,int\n',
                    'second,bar,str\n',
                    'third,4.4,float\n',
                ])

            data = json.loads(create_analysis_settings_json(d))

            self.assertEqual(data['model_settings'], {
                'first': 3,
                'second': 'bar',
                'third': 4.4,
            })

    def test_gul_settings_contains_data___data_is_stored_in_gul_summaries(self):
        with TemporaryDirectory() as d:
            self.create_fake_directory(d)

            with io.open(os.path.join(d, GUL_SUMMARIES_FILE), 'w', encoding='utf-8') as f:
                f.writelines([
                    '1,leccalc_foo,TRUE\n',
                    '1,leccalc_bar,FALSE\n',
                    '1,another,FALSE\n',
                    '2,leccalc_boo,FALSE\n',
                    '2,leccalc_far,TRUE\n',
                    '2,different,TRUE\n',
                ])

            data = json.loads(create_analysis_settings_json(d))

            self.assertEqual(data['gul_summaries'], [
                {'id': 1, 'another': False, 'leccalc': {'leccalc_foo': True, 'leccalc_bar': False}},
                {'id': 2, 'different': True, 'leccalc': {'leccalc_boo': False, 'leccalc_far': True}},
            ])

    def test_il_settings_contains_data___data_is_stored_in_il_summaries(self):
        with TemporaryDirectory() as d:
            self.create_fake_directory(d)

            with io.open(os.path.join(d, IL_SUMMARIES_FILE), 'w', encoding='utf-8') as f:
                f.writelines([
                    '1,leccalc_foo,TRUE\n',
                    '1,leccalc_bar,FALSE\n',
                    '1,another,FALSE\n',
                    '2,leccalc_boo,FALSE\n',
                    '2,leccalc_far,TRUE\n',
                    '2,different,TRUE\n',
                ])

            data = json.loads(create_analysis_settings_json(d))

            self.assertEqual(data['il_summaries'], [
                {'id': 1, 'another': False, 'leccalc': {'leccalc_foo': True, 'leccalc_bar': False}},
                {'id': 2, 'different': True, 'leccalc': {'leccalc_boo': False, 'leccalc_far': True}},
            ])



from oasislmf.exposures.manager import OasisExposuresManager


def fake_model(supplier='supplier', model='model', version='version', resources=None):
    return OasisExposuresManager().create(supplier, model, version, resources=resources)



from __future__ import unicode_literals

import glob
import tarfile
from tempfile import NamedTemporaryFile

import six
from chainmap import ChainMap
from itertools import chain
from backports.tempfile import TemporaryDirectory
from unittest import TestCase

import os
import io
import subprocess

from copy import deepcopy
from hypothesis import given
from hypothesis.strategies import sampled_from, lists
from mock import patch, Mock
from pathlib2 import Path

from oasislmf.model_execution.files import GUL_INPUT_FILES, OPTIONAL_INPUT_FILES, IL_INPUT_FILES, TAR_FILE, INPUT_FILES
from oasislmf.model_execution.bin import create_binary_files, create_binary_tar_file, check_conversion_tools, \
    check_inputs_directory, prepare_model_run_directory, prepare_model_run_inputs, cleanup_bin_directory, \
    check_binary_tar_file
from oasislmf.utils.exceptions import OasisException

ECHO_CONVERSION_INPUT_FILES = {k: ChainMap({'conversion_tool': 'echo'}, v) for k, v in INPUT_FILES.items()}


def standard_input_files(min_size=0):
    return lists(
        sampled_from([target['name'] for target in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(OPTIONAL_INPUT_FILES))]),
        min_size=min_size,
        unique=True,
    )


def il_input_files(min_size=0):
    return lists(
        sampled_from([target['name'] for target in six.itervalues(IL_INPUT_FILES)]),
        min_size=min_size,
        unique=True,
    )


def tar_file_targets(min_size=0):
    return lists(
        sampled_from([target['name'] + '.bin' for target in six.itervalues(INPUT_FILES)]),
        min_size=min_size,
        unique=True,
    )


class CreateBinaryFiles(TestCase):
    def test_directory_only_contains_excluded_files___tar_is_empty(self):
        with TemporaryDirectory() as csv_dir, TemporaryDirectory() as bin_dir:
            with io.open(os.path.join(csv_dir, 'another_file'), 'w', encoding='utf-8') as f:
                f.write('file data')

            create_binary_files(csv_dir, bin_dir)

            self.assertEqual(0, len(glob.glob(os.path.join(csv_dir, '*.bin'))))

    @given(standard_input_files(min_size=1), il_input_files(min_size=1))
    def test_contains_il_and_standard_files_but_do_il_is_false___il_files_are_excluded(self, standard, il):
        with patch('oasislmf.model_execution.bin.INPUT_FILES', ECHO_CONVERSION_INPUT_FILES), TemporaryDirectory() as csv_dir, TemporaryDirectory() as bin_dir:
            for target in chain(standard, il):
                with io.open(os.path.join(csv_dir, target + '.csv'), 'w', encoding='utf-8') as f:
                    f.write(target)

            create_binary_files(csv_dir, bin_dir, do_il=False)

            self.assertEqual(len(standard), len(glob.glob(os.path.join(bin_dir, '*.bin'))))
            for filename in (f + '.bin' for f in standard):
                self.assertTrue(os.path.exists(os.path.join(bin_dir, filename)))

    @given(standard_input_files(min_size=1), il_input_files(min_size=1))
    def test_contains_il_and_standard_files_but_do_il_is_true___all_files_are_included(self, standard, il):
        with patch('oasislmf.model_execution.bin.INPUT_FILES', ECHO_CONVERSION_INPUT_FILES), TemporaryDirectory() as csv_dir, TemporaryDirectory() as bin_dir:
            for target in chain(standard, il):
                with io.open(os.path.join(csv_dir, target + '.csv'), 'w', encoding='utf-8') as f:
                    f.write(target)

            create_binary_files(csv_dir, bin_dir, do_il=True)

            self.assertEqual(len(standard) + len(il), len(glob.glob(os.path.join(bin_dir, '*.bin'))))
            for filename in (f + '.bin' for f in chain(standard, il)):
                self.assertTrue(os.path.exists(os.path.join(bin_dir, filename)))

    def test_subprocess_raises___oasis_exception_is_raised(self):
        with TemporaryDirectory() as csv_dir, TemporaryDirectory() as bin_dir:
            Path(os.path.join(csv_dir, 'events.csv')).touch()

            with patch('oasislmf.model_execution.bin.subprocess.check_call', Mock(side_effect=subprocess.CalledProcessError(1, ''))):
                with self.assertRaises(OasisException):
                    create_binary_files(csv_dir, bin_dir, do_il=True)


class CreateBinaryTarFile(TestCase):
    def test_directory_only_contains_excluded_files___tar_is_empty(self):
        with TemporaryDirectory() as d:
            with io.open(os.path.join(d, 'another_file'), 'w', encoding='utf-8') as f:
                f.write('file data')

            create_binary_tar_file(d)

            with tarfile.open(os.path.join(d, TAR_FILE), 'r:gz', encoding='utf-8') as tar:
                self.assertEqual(0, len(tar.getnames()))

    @given(tar_file_targets(min_size=1))
    def test_directory_contains_some_target_files___target_files_are_included(self, targets):
        with TemporaryDirectory() as d:
            for target in targets:
                with io.open(os.path.join(d, target), 'w', encoding='utf-8') as f:
                    f.write(target)

            create_binary_tar_file(d)

            with tarfile.open(os.path.join(d, TAR_FILE), 'r:gz', encoding='utf-8') as tar:
                self.assertEqual(len(targets), len(tar.getnames()))
                self.assertEqual(set(targets), set(tar.getnames()))


class CheckConversionTools(TestCase):
    def test_do_il_is_false_il_tools_are_missing___result_is_true(self):
        existing_conversions = deepcopy(INPUT_FILES)
        for value in six.itervalues(existing_conversions):
            if value['type'] == 'il':
                value['conversion_tool'] = 'missing_executable'
            else:
                value['conversion_tool'] = 'python'

        with patch('oasislmf.model_execution.bin.INPUT_FILES', existing_conversions):
            self.assertTrue(check_conversion_tools())

    def test_do_il_is_false_il_tools_are_present_but_non_il_are_missing___errors_is_raised(self):
        existing_conversions = deepcopy(INPUT_FILES)
        for value in six.itervalues(existing_conversions):
            if value['type'] == 'il':
                value['conversion_tool'] = 'pytohn'
            else:
                value['conversion_tool'] = 'missing_executable'

        with patch('oasislmf.model_execution.bin.INPUT_FILES', existing_conversions):
            with self.assertRaises(OasisException):
                check_conversion_tools()

    def test_do_il_is_true_il_tools_are_missing___error_is_raised(self):
        existing_conversions = deepcopy(INPUT_FILES)
        for value in six.itervalues(existing_conversions):
            if value['type'] == 'il':
                value['conversion_tool'] = 'missing_executable'
            else:
                value['conversion_tool'] = 'python'

        with patch('oasislmf.model_execution.bin.INPUT_FILES', existing_conversions):
            with self.assertRaises(OasisException):
                check_conversion_tools(do_il=True)

    def test_do_il_is_true_non_il_are_missing___errror_is_raised(self):
        existing_conversions = deepcopy(INPUT_FILES)
        for value in six.itervalues(existing_conversions):
            if value['type'] == 'il':
                value['conversion_tool'] = 'pytohn'
            else:
                value['conversion_tool'] = 'missing_executable'

        with patch('oasislmf.model_execution.bin.INPUT_FILES', existing_conversions):
            with self.assertRaises(OasisException):
                check_conversion_tools(do_il=True)

    def test_do_il_is_true_conversion_tools_all_exist___result_is_true(self):
        existing_conversions = deepcopy(INPUT_FILES)
        for value in existing_conversions.values():
            value['conversion_tool'] = 'python'

        with patch('oasislmf.model_execution.bin.INPUT_FILES', existing_conversions):
            self.assertTrue(check_conversion_tools(do_il=True))


class CheckInputDirectory(TestCase):
    def test_tar_file_already_exists___exception_is_raised(self):
        with TemporaryDirectory() as d:
            Path(os.path.join(d, TAR_FILE)).touch()
            with self.assertRaises(OasisException):
                check_inputs_directory(d, False)

    @given(il_input_files())
    def test_do_il_is_false_non_il_input_files_are_missing__exception_is_raised(self, il_files):
        with TemporaryDirectory() as d:
            for p in il_files:
                Path(os.path.join(d, p + '.csv')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, False)

    def test_do_is_is_false_non_il_input_files_are_present___no_exception_is_raised(self):
        with TemporaryDirectory() as d:
            for input_file in six.itervalues(GUL_INPUT_FILES):
                Path(os.path.join(d, input_file['name'] + '.csv')).touch()

            try:
                check_inputs_directory(d, False)
            except Exception as e:
                self.fail('Exception was raised {}: {}'.format(type(e), e))

    def test_do_il_is_true_all_input_files_are_missing__exception_is_raised(self):
        with TemporaryDirectory() as d:
            with self.assertRaises(OasisException):
                check_inputs_directory(d, True)

    def test_do_il_is_true_gul_input_files_are_missing__exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in six.itervalues(IL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, True)

    def test_do_il_is_true_il_input_files_are_missing__exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in six.itervalues(GUL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, True)

    def test_do_il_is_true_all_input_files_are_present___no_exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            try:
                check_inputs_directory(d, True)
            except Exception as e:
                self.fail('Exception was raised {}: {}'.format(type(e), e))

    def test_do_il_is_false_il_bin_files_are_present___no_exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(IL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            try:
                check_inputs_directory(d, False)
            except Exception as e:
                self.fail('Exception was raised {}: {}'.format(type(e), e))

    def test_do_il_is_false_gul_bin_files_are_present___exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(GUL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, False)

    def test_do_il_is_true_gul_bin_files_are_present___exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(GUL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, True)

    def test_do_il_is_true_il_bin_files_are_present___exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(IL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            with self.assertRaises(OasisException):
                check_inputs_directory(d, True)

    def test_do_il_is_true_no_bin_files_are_present___no_exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES)):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(IL_INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            try:
                check_inputs_directory(d, False)
            except Exception as e:
                self.fail('Exception was raised {}: {}'.format(type(e), e))

    def test_do_il_is_true_bin_files_are_present_but_check_bin_files_are_true___no_exception_is_raised(self):
        with TemporaryDirectory() as d:
            for p in six.itervalues(INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.csv')).touch()

            for p in six.itervalues(INPUT_FILES):
                Path(os.path.join(d, p['name'] + '.bin')).touch()

            try:
                check_inputs_directory(d, do_il=True, check_binaries=False)
            except Exception as e:
                self.fail('Exception was raised {}: {}'.format(type(e), e))


class PrepareModelRunDirectory(TestCase):
    def test_directory_is_empty___child_directories_are_created(self):
        with TemporaryDirectory() as d:
            prepare_model_run_directory(d)

            self.assertTrue(os.path.exists(os.path.join(d, 'fifo')))
            self.assertTrue(os.path.exists(os.path.join(d, 'input')))
            self.assertTrue(os.path.exists(os.path.join(d, 'input', 'csv')))
            self.assertTrue(os.path.exists(os.path.join(d, 'output')))
            self.assertTrue(os.path.exists(os.path.join(d, 'static')))
            self.assertTrue(os.path.exists(os.path.join(d, 'work')))

    def test_directory_has_some_exisitng_directories___other_child_directories_are_created(self):
        with TemporaryDirectory() as d:
            os.mkdir(os.path.join(d, 'fifo'))
            os.mkdir(os.path.join(d, 'input'))

            prepare_model_run_directory(d)

            self.assertTrue(os.path.exists(os.path.join(d, 'fifo')))
            self.assertTrue(os.path.exists(os.path.join(d, 'input')))
            self.assertTrue(os.path.exists(os.path.join(d, 'input', 'csv')))
            self.assertTrue(os.path.exists(os.path.join(d, 'output')))
            self.assertTrue(os.path.exists(os.path.join(d, 'static')))
            self.assertTrue(os.path.exists(os.path.join(d, 'work')))

    def test_input_directory_is_supplied___input_files_are_copied_to_input_csv(self):
        with TemporaryDirectory() as output_path, TemporaryDirectory() as input_path:
            Path(os.path.join(input_path, 'a_file.csv')).touch()

            prepare_model_run_directory(output_path, oasis_files_src_path=input_path)

            self.assertTrue(os.path.exists(os.path.join(output_path, 'input', 'csv', 'a_file.csv')))

    def test_settings_file_is_supplied___settings_file_is_copied_into_run_dir(self):
        with TemporaryDirectory() as output_path, NamedTemporaryFile('w') as input_file:
            input_file.write('conf stuff')
            input_file.flush()

            prepare_model_run_directory(output_path, analysis_settings_json_src_file_path=input_file.name)

            with io.open(os.path.join(output_path, 'analysis_settings.json'), encoding='utf-8') as output_conf:
                self.assertEqual('conf stuff', output_conf.read())

    def test_model_data_src_is_supplied___symlink_to_output_dir_static_is_created(self):
        with TemporaryDirectory() as output_path, TemporaryDirectory() as input_path:
            Path(os.path.join(input_path, 'linked_file')).touch()

            prepare_model_run_directory(output_path, model_data_src_path=input_path)

            self.assertTrue(os.path.exists(os.path.join(output_path, 'static', 'linked_file')))

    def test_model_data_src_is_supplied_sym_link_raises___input_is_copied_from_static(self):
        with TemporaryDirectory() as output_path, TemporaryDirectory() as input_path:
            Path(os.path.join(input_path, 'linked_file')).touch()

            with patch('os.symlink', Mock(side_effect=OSError())):
                prepare_model_run_directory(output_path, model_data_src_path=input_path)

            self.assertTrue(os.path.exists(os.path.join(output_path, 'static', 'linked_file')))

    def test_inputs_archive_is_supplied___archive_is_extracted_into_inputs(self):
        with TemporaryDirectory() as output_path, TemporaryDirectory() as input_path:
            tar_path = os.path.join(input_path, 'archive.tar')

            with tarfile.open(tar_path, 'w', encoding='utf-8') as tar:
                archived_file_path = Path(input_path, 'archived_file')
                archived_file_path.touch()
                tar.add(str(archived_file_path), arcname='archived_file')

            prepare_model_run_directory(output_path, inputs_archive=tar_path)

            self.assertTrue(Path(output_path, 'input', 'archived_file').exists())


class PrepareModelRunInputs(TestCase):
    def make_fake_bins(self, d):
        os.mkdir(os.path.join(d, 'input'))
        os.mkdir(os.path.join(d, 'static'))
        Path(os.path.join(d, 'static', 'events.bin')).touch()
        Path(os.path.join(d, 'static', 'returnperiods.bin')).touch()
        Path(os.path.join(d, 'static', 'occurrence.bin')).touch()
        Path(os.path.join(d, 'static', 'periods.bin')).touch()

    def test_prepare_input_bin_raises___oasis_exception_is_raised(self):
        with patch('oasislmf.model_execution.bin._prepare_input_bin', Mock(side_effect=OSError('os error'))):
            with self.assertRaises(OasisException):
                prepare_model_run_inputs({}, 'some_dir')

    def test_events_bin_already_exists___existing_bin_is_uncahnged(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'input', 'events.bin'), 'w', encoding='utf-8') as events_file:
                events_file.write('events bin')
                events_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'events.bin'), 'r', encoding='utf-8') as new_events_file:
                self.assertEqual('events bin', new_events_file.read())

    def test_events_bin_doesnt_not_exist_event_set_isnt_specified___bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'events.bin'), 'w', encoding='utf-8') as events_file:
                events_file.write('events bin')
                events_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'events.bin'), 'r', encoding='utf-8') as new_events_file:
                self.assertEqual('events bin', new_events_file.read())

    def test_events_bin_doesnt_not_exist_event_set_is_specified___event_set_specific_bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'events_from_set.bin'), 'w', encoding='utf-8') as events_file:
                events_file.write('events from set bin')
                events_file.flush()

                prepare_model_run_inputs({'model_settings': {'event_set': 'from set'}}, d)

            with io.open(os.path.join(d, 'input', 'events.bin'), 'r', encoding='utf-8') as new_events_file:
                self.assertEqual('events from set bin', new_events_file.read())

    def test_no_events_bin_exists___oasis_exception_is_raised(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)
            os.remove(os.path.join(d, 'static', 'events.bin'))

            with self.assertRaises(OasisException):
                prepare_model_run_inputs({}, d)

    def test_returnperiods_bin_already_exists___existing_bin_is_uncahnged(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'input', 'returnperiods.bin'), 'w', encoding='utf-8') as returnperiods_file:
                returnperiods_file.write('returnperiods bin')
                returnperiods_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'returnperiods.bin'), 'r', encoding='utf-8') as new_returnperiods_file:
                self.assertEqual('returnperiods bin', new_returnperiods_file.read())

    def test_returnperiods_bin_doesnt_not_exist_event_set_isnt_specified___bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'returnperiods.bin'), 'w', encoding='utf-8') as returnperiods_file:
                returnperiods_file.write('returnperiods bin')
                returnperiods_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'returnperiods.bin'), 'r', encoding='utf-8') as new_returnperiods_file:
                self.assertEqual('returnperiods bin', new_returnperiods_file.read())

    def test_no_returnperiods_bin_exists___oasis_exception_is_raised(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)
            os.remove(os.path.join(d, 'static', 'returnperiods.bin'))

            with self.assertRaises(OasisException):
                prepare_model_run_inputs({}, d)

    def test_occurrence_bin_already_exists___existing_bin_is_uncahnged(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'input', 'occurrence.bin'), 'w', encoding='utf-8') as occurrence_file:
                occurrence_file.write('occurrence bin')
                occurrence_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'occurrence.bin'), 'r', encoding='utf-8') as new_occurrence_file:
                self.assertEqual('occurrence bin', new_occurrence_file.read())

    def test_occurrence_bin_doesnt_not_exist_event_set_isnt_specified___bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'occurrence.bin'), 'w', encoding='utf-8') as occurrence_file:
                occurrence_file.write('occurrence bin')
                occurrence_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'occurrence.bin'), 'r', encoding='utf-8') as new_occurrence_file:
                self.assertEqual('occurrence bin', new_occurrence_file.read())

    def test_occurrence_bin_doesnt_not_exist_event_set_is_specified___event_occurrence_id_specific_bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'occurrence_occurrence_id.bin'), 'w', encoding='utf-8') as occurrence_file:
                occurrence_file.write('occurrence occurrence id bin')
                occurrence_file.flush()

                prepare_model_run_inputs({'model_settings': {'event_occurrence_id': 'occurrence id'}}, d)

            with io.open(os.path.join(d, 'input', 'occurrence.bin'), 'r', encoding='utf-8') as new_occurrence_file:
                self.assertEqual('occurrence occurrence id bin', new_occurrence_file.read())

    def test_no_occurrence_bin_exists___oasis_exception_is_raised(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)
            os.remove(os.path.join(d, 'static', 'occurrence.bin'))

            with self.assertRaises(OasisException):
                prepare_model_run_inputs({}, d)

    def test_periods_bin_already_exists___existing_bin_is_uncahnged(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'input', 'periods.bin'), 'w', encoding='utf-8') as periods_file:
                periods_file.write('periods bin')
                periods_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'periods.bin'), 'r', encoding='utf-8') as new_periods_file:
                self.assertEqual('periods bin', new_periods_file.read())

    def test_periods_bin_doesnt_not_exist_event_set_isnt_specified___bin_is_copied_from_static(self):
        with TemporaryDirectory() as d:
            self.make_fake_bins(d)

            with io.open(os.path.join(d, 'static', 'periods.bin'), 'w', encoding='utf-8') as periods_file:
                periods_file.write('periods bin')
                periods_file.flush()

                prepare_model_run_inputs({}, d)

            with io.open(os.path.join(d, 'input', 'periods.bin'), 'r', encoding='utf-8') as new_periods_file:
                self.assertEqual('periods bin', new_periods_file.read())


class CleanBinDirectory(TestCase):
    def test_output_and_bin_input_files_are_removed(self):
        with TemporaryDirectory() as d:
            Path(os.path.join(d, TAR_FILE)).touch()

            for f in six.iterkeys(INPUT_FILES):
                Path(os.path.join(d, f + '.bin')).touch()

            cleanup_bin_directory(d)

            self.assertFalse(os.path.exists(os.path.join(d, TAR_FILE)))
            for f in six.iterkeys(INPUT_FILES):
                self.assertFalse(os.path.exists(os.path.join(d, f + '.bin')))


class CheckBinTarFile(TestCase):
    def test_all_il_files_are_missing_check_il_is_false___result_is_true(self):
        with TemporaryDirectory() as d:
            tar_file_name = os.path.join(d, 'exposures.tar')

            with tarfile.open(tar_file_name, 'w', encoding='utf-8') as tar:
                for f in six.itervalues(GUL_INPUT_FILES):
                    Path(os.path.join(d, '{}.bin'.format(f['name']))).touch()

                tar.add(d, arcname='/')

            self.assertTrue(check_binary_tar_file(tar_file_name))

    def test_all_files_are_present_check_il_is_true___result_is_true(self):
        with TemporaryDirectory() as d:
            tar_file_name = os.path.join(d, 'exposures.tar')

            with tarfile.open(tar_file_name, 'w', encoding='utf-8') as tar:
                for f in six.itervalues(INPUT_FILES):
                    Path(os.path.join(d, '{}.bin'.format(f['name']))).touch()

                tar.add(d, arcname='/')

            self.assertTrue(check_binary_tar_file(tar_file_name, check_il=True))

    @given(
        lists(sampled_from([f['name'] for f in chain(six.itervalues(GUL_INPUT_FILES), six.itervalues(IL_INPUT_FILES))]), min_size=1, unique=True)
    )
    def test_some_files_are_missing_check_il_is_true___error_is_raised(self, missing):
        with TemporaryDirectory() as d:
            tar_file_name = os.path.join(d, 'exposures.tar')

            with tarfile.open(tar_file_name, 'w', encoding='utf-8') as tar:
                for f in six.itervalues(INPUT_FILES):
                    if f['name'] in missing:
                        continue

                    Path(os.path.join(d, '{}.bin'.format(f['name']))).touch()

                tar.add(d, arcname='/')

            with self.assertRaises(OasisException):
                check_binary_tar_file(tar_file_name, check_il=True)



import json
from collections import Counter
from tempfile import NamedTemporaryFile
from unittest import TestCase

import os

import shutil
import six
from hypothesis import given
from hypothesis.strategies import sampled_from, text, dictionaries, booleans, integers
from mock import Mock, ANY, patch
from backports.tempfile import mkdtemp, TemporaryDirectory
from pathlib2 import Path

from oasislmf.api_client.client import OasisAPIClient
from oasislmf.cmd import RootCmd
from oasislmf.cmd.test import TestModelApiCmd
from oasislmf.utils.exceptions import OasisException


class TestModelApiCmdLoadAnalysisSettingsJson(TestCase):
    def test_do_il_is_true___result_has_input_dict_and_do_il_is_true(self):
        conf = {
            'analysis_settings': {
                'il_output': True,
                'foo': 'bar',
            }
        }

        with NamedTemporaryFile('w') as f:
            json.dump(conf, f)
            f.flush()

            res_conf, do_il = TestModelApiCmd().load_analysis_settings_json(f.name)

            self.assertEqual(conf, res_conf)
            self.assertTrue(do_il)

    def test_do_il_is_false___result_has_input_dict_and_do_il_is_false(self):
        conf = {
            'analysis_settings': {
                'il_output': False,
                'foo': 'bar',
            }
        }

        with NamedTemporaryFile('w') as f:
            json.dump(conf, f)
            f.flush()

            res_conf, do_il = TestModelApiCmd().load_analysis_settings_json(f.name)

            self.assertEqual(conf, res_conf)
            self.assertFalse(do_il)

    @given(sampled_from(['true', 'TRUE', 'True']))
    def test_do_il_is_string_true___result_has_input_dict_and_do_il_is_true(self, do_il_in):
        conf = {
            'analysis_settings': {
                'il_output': do_il_in,
                'foo': 'bar',
            }
        }

        with NamedTemporaryFile('w') as f:
            json.dump(conf, f)
            f.flush()

            res_conf, do_il = TestModelApiCmd().load_analysis_settings_json(f.name)

            self.assertEqual(conf, res_conf)
            self.assertTrue(do_il)

    @given(sampled_from(['false', 'FALSE', 'False']))
    def test_do_il_is_string_false___result_has_input_dict_and_do_il_is_false(self, do_il_in):
        conf = {
            'analysis_settings': {
                'il_output': do_il_in,
                'foo': 'bar',
            }
        }

        with NamedTemporaryFile('w') as f:
            json.dump(conf, f)
            f.flush()

            res_conf, do_il = TestModelApiCmd().load_analysis_settings_json(f.name)

            self.assertEqual(conf, res_conf)
            self.assertFalse(do_il)


class TestModelApiCmdRunAnalysis(TestCase):
    @given(text(), text(), dictionaries(text(), text()), booleans(), integers(min_value=0, max_value=5), integers(min_value=0, max_value=5))
    def test_no_errors_are_raised___completed_is_incremented(self, input_dir, output_dir, settings, do_il, initial_complete, initial_failed):
        client = OasisAPIClient('http://localhost:8001')
        client.upload_inputs_from_directory = Mock(return_value='input_location')
        client.run_analysis_and_poll = Mock()

        counter = Counter({
            'completed': initial_complete,
            'failed': initial_failed,
        })

        TestModelApiCmd().run_analysis((client, input_dir, output_dir, settings, do_il, counter))

        self.assertEqual(initial_complete + 1, counter['completed'])
        self.assertEqual(initial_failed, counter['failed'])
        client.upload_inputs_from_directory.assert_called_once_with(input_dir, bin_directory=ANY, do_il=do_il, do_build=True)
        client.run_analysis_and_poll.assert_called_once_with(settings, 'input_location', output_dir)

    @given(text(), text(), dictionaries(text(), text()), booleans(), integers(min_value=0, max_value=5), integers(min_value=0, max_value=5))
    def test_uploading_raises_an_error___failed_counter_is_incremented(self, input_dir, output_dir, settings, do_il, initial_complete, initial_failed):
        client = OasisAPIClient('http://localhost:8001')
        client.upload_inputs_from_directory = Mock(side_effect=OasisException())
        client.run_analysis_and_poll = Mock()

        counter = Counter({
            'completed': initial_complete,
            'failed': initial_failed,
        })

        TestModelApiCmd().run_analysis((client, input_dir, output_dir, settings, do_il, counter))

        self.assertEqual(initial_complete, counter['completed'])
        self.assertEqual(initial_failed + 1, counter['failed'])

    @given(text(), text(), dictionaries(text(), text()), booleans(), integers(min_value=0, max_value=5), integers(min_value=0, max_value=5))
    def test_run_and_poll_raises_an_error___failed_counter_is_incremented(self, input_dir, output_dir, settings, do_il, initial_complete, initial_failed):
        client = OasisAPIClient('http://localhost:8001')
        client.upload_inputs_from_directory = Mock()
        client.run_analysis_and_poll = Mock(side_effect=OasisException())

        counter = Counter({
            'completed': initial_complete,
            'failed': initial_failed,
        })

        TestModelApiCmd().run_analysis((client, input_dir, output_dir, settings, do_il, counter))

        self.assertEqual(initial_complete, counter['completed'])
        self.assertEqual(initial_failed + 1, counter['failed'])


class TestModelApiCmdRun(TestCase):
    def setUp(self):
        self.directory = mkdtemp()
        self._orig_cwd = os.getcwd()
        self.create_working_directory(self.directory)

    def tearDown(self):
        os.chdir(self._orig_cwd)
        shutil.rmtree(self.directory)

    def create_working_directory(self, d):
        os.mkdir(os.path.join(d, 'input'))
        Path(os.path.join(d, 'analysis_settings.json')).touch()

        os.chdir(d)

    def get_command(self, api_server_url='http://localhost:8001', analysis_directory=None, extras=None):
        kwargs = {}

        if analysis_directory:
            kwargs['input-directory'] = os.path.join(analysis_directory, 'input')
            kwargs['output-directory'] = os.path.join(analysis_directory, 'output')
            kwargs['analysis-settings-file'] = os.path.join(analysis_directory, 'analysis_settings.json')

        kwargs.update(extras or {})
        kwargs_str = ' '.join('--{} {}'.format(k, v) for k, v in six.iteritems(kwargs))

        return RootCmd(argv='test model-api {} {}'.format(kwargs_str, api_server_url).split())

    def test_api_server_is_not_supplied___parsing_args_raises_an_error(self):
        cmd = self.get_command(api_server_url='')

        with self.assertRaises(SystemExit):
            cmd.parse_args()

    def test_input_directory_is_not_supplied___default_is_abs_path_to_input_in_cwd(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(args.input_directory, os.path.abspath('input'))

    def test_input_directory_is_supplied___path_is_abs_path_to_supplied_dir(self):
        with TemporaryDirectory() as input_dir:
            cmd = self.get_command(extras={'input-directory': input_dir})

            args = cmd.parse_args()

            self.assertEqual(args.input_directory, os.path.abspath(input_dir))

    def test_output_directory_is_not_supplied___default_is_abs_path_to_output_in_cwd(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(args.output_directory, os.path.abspath('output'))

    def test_output_directory_is_supplied___path_is_abs_path_to_supplied_dir(self):
        with TemporaryDirectory() as output_dir:
            cmd = self.get_command(extras={'output-directory': output_dir})

            args = cmd.parse_args()

            self.assertEqual(args.output_directory, os.path.abspath(output_dir))

    def test_analysis_settings_is_not_supplied___default_is_abs_path_to_analysis_settings_in_cwd(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(args.analysis_settings_file, os.path.abspath('analysis_settings.json'))

    def test_analysis_settings_is_supplied___path_is_abs_path_to_supplied_file(self):
        with TemporaryDirectory() as settings_path:
            settings_path = os.path.join(settings_path, 'analysis_settings.json')
            Path(settings_path).touch()

            cmd = self.get_command(extras={'analysis-settings-file': settings_path})

            args = cmd.parse_args()

            self.assertEqual(args.analysis_settings_file, os.path.abspath(settings_path))

    def test_num_analyses_is_not_supplied___default_1(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(1, args.num_analyses)

    def test_num_analyses_is_supplied___num_analyses_is_correct(self):
        cmd = self.get_command(extras={'num-analyses': 6})

        args = cmd.parse_args()

        self.assertEqual(6, args.num_analyses)

    def test_health_check_attempts_is_not_supplied___default_1(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(1, args.health_check_attempts)

    def test_health_check_attempts_is_supplied___health_check_attempts_is_correct(self):
        cmd = self.get_command(extras={'health-check-attempts': 6})

        args = cmd.parse_args()

        self.assertEqual(6, args.health_check_attempts)

    def test_input_directory_does_not_exist___error_is_raised(self):
        shutil.rmtree(os.path.join(self.directory, 'input'))

        cmd = self.get_command(analysis_directory=self.directory)
        cmd._logger = Mock()

        res = cmd.run()

        self.assertEqual(1, res)
        cmd.logger.error.assert_called_once_with('Input directory does not exist: {}'.format(os.path.join(self.directory, 'input')))

    def test_analysis_settings_file_does_not_exist___error_is_raised(self):
        os.remove(os.path.join(self.directory, 'analysis_settings.json'))

        cmd = self.get_command(analysis_directory=self.directory)
        cmd._logger = Mock()

        res = cmd.run()

        self.assertEqual(1, res)
        cmd.logger.error.assert_called_once_with('Analysis settings does not exist: {}'.format(os.path.join(self.directory, 'analysis_settings.json')))

    @given(integers(min_value=1, max_value=5))
    def test_client_health_check_fails___error_is_written_to_the_log(self, health_check_attempts):
        with patch('oasislmf.api_client.client.OasisAPIClient.health_check', Mock(return_value=False)) as health_check_mock:
            cmd = self.get_command(analysis_directory=self.directory, extras={'health-check-attempts': health_check_attempts})
            cmd._logger = Mock()

            res = cmd.run()

            self.assertEqual(1, res)
            cmd.logger.error.assert_called_once_with('Health check failed for http://localhost:8001.')
            health_check_mock.assert_called_once_with(health_check_attempts)

    @given(integers(min_value=1, max_value=5), booleans(), dictionaries(text(), text()))
    def test_validation_is_successful___threadpool_is_started_for_each_analysis(self, num_analyses, do_il, settings):
        pool_mock_object = Mock()

        with patch('oasislmf.api_client.client.OasisAPIClient.health_check', Mock(return_value=True)), \
                patch('oasislmf.cmd.test.TestModelApiCmd.load_analysis_settings_json', Mock(return_value=(settings, do_il))), \
                patch('oasislmf.cmd.test.ThreadPool', Mock(return_value=pool_mock_object)) as pool_mock:
            cmd = self.get_command(analysis_directory=self.directory, extras={'num-analyses': num_analyses})
            cmd._logger = Mock()

            res = cmd.run()

            pool_mock.assert_called_once_with(processes=num_analyses)

            self.assertEqual(0, res)
            self.assertEqual(1, pool_mock_object.map.call_count)
            fn = pool_mock_object.map.call_args[0][0]
            args = list(pool_mock_object.map.call_args[0][1])

            self.assertEqual(fn.__name__, 'run_analysis')
            self.assertIsInstance(fn.__self__, TestModelApiCmd)
            self.assertEqual(args, [(ANY, cmd.args.input_directory, cmd.args.output_directory, settings, do_il, ANY)] * num_analyses)



# -*- coding: utf-8 -*-

__all__ = [
    'OasisFilesPipeline'
]


class OasisFilesPipeline(object):
    def __init__(
        self,
        model_key=None,
        source_exposures_file_path=None,
        canonical_exposures_file_path=None,
        model_exposures_file_path=None,
        keys_file_path=None,
        keys_errors_file_path=None
    ):
        self._model_key = model_key
        self._source_exposures_file_path = source_exposures_file_path
        self._canonical_exposures_file_path = canonical_exposures_file_path
        self._model_exposures_file_path = model_exposures_file_path
        self._keys_file_path = keys_file_path
        self._keys_errors_file_path = keys_errors_file_path

        self._items_file_path = None
        self._coverages_file_path = None
        self._gulsummaryxref_file_path = None

        self._oasis_files = {
            'items': self._items_file_path,
            'coverages': self._coverages_file_path,
            'gulsummaryxref': self._gulsummaryxref_file_path
        }

        self._file_paths = (
            'source_exposures_file_path',
            'canonical_exposures_file_path',
            'model_exposures_file_path',
            'keys_file_path',
            'keys_errors_file_path',
            'items_file_path',
            'coverages_file_path',
            'gulsummaryxref_file_path',
        )

    def __str__(self):
        return '{}: {}'.format(self.__repr__(), self.model_key)

    def __repr__(self):
        return '{}: {}'.format(self.__class__, self.__dict__)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else '...')

    @property
    def model_key(self):
        """
        Model key property - getter only.

            :getter: Gets the key of model to which the pipeline is attached.
        """
        return self._model_key

    @property
    def source_exposures_file_path(self):
        """
        Source exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._source_exposures_file_path

    @source_exposures_file_path.setter
    def source_exposures_file_path(self, p):
        self._source_exposures_file_path = p

    @property
    def canonical_exposures_file_path(self):
        """
        Canonical exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._canonical_exposures_file_path

    @canonical_exposures_file_path.setter
    def canonical_exposures_file_path(self, p):
        self._canonical_exposures_file_path = p

    @property
    def model_exposures_file_path(self):
        """
        Model exposures file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._model_exposures_file_path

    @model_exposures_file_path.setter
    def model_exposures_file_path(self, p):
        self._model_exposures_file_path = p

    @property
    def keys_file_path(self):
        """
        Oasis keys file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._keys_file_path

    @keys_file_path.setter
    def keys_file_path(self, p):
        self._keys_file_path = p

    @property
    def keys_errors_file_path(self):
        """
        Oasis keys error file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._keys_errors_file_path

    @keys_errors_file_path.setter
    def keys_errors_file_path(self, p):
        self._keys_errors_file_path = p

    @property
    def items_file_path(self):
        """
        Oasis items file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._items_file_path

    @items_file_path.setter
    def items_file_path(self, p):
        self._items_file_path = self.oasis_files['items'] = p

    @property
    def coverages_file_path(self):
        """
        Oasis coverages file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._coverages_file_path

    @coverages_file_path.setter
    def coverages_file_path(self, p):
        self._coverages_file_path = self.oasis_files['coverages'] = p

    @property
    def gulsummaryxref_file_path(self):
        """
        GUL summary file path property.

            :getter: Gets the file path
            :setter: Sets the current file path to the specified file path
        """
        return self._gulsummaryxref_file_path

    @gulsummaryxref_file_path.setter
    def gulsummaryxref_file_path(self, p):
        self._gulsummaryxref_file_path = self.oasis_files['gulsummaryxref'] = p

    @property
    def oasis_files(self):
        """
        Oasis files set property - getter only.

            :getter: Gets the complete set of paths of the generated Oasis
            files, including  ``items.csv``, ``coverages.csv``, `gulsummaryxref.csv`.
        """
        return self._oasis_files

    def clear(self):
        """
        Clears all file path attributes in the pipeline.
        """
        [setattr(self, p, None) for p in self._file_paths]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

__all__ = [
    'OasisModel',
    'OasisModelFactory'
]


class OasisModel(object):
    """
    A simple object representation of Oasis models and their resources - an
    Oasis model is identified by a triple: a specific supplier ID, model ID
    and model version. The constructor requires these three arguments
    for creating a new Oasis model object. Each model object also has a
    resources dictionary that can be used to "attach" any resources by clients,
    e.g. a lookup service instance, a transforms files pipeline, validation
    and transformation files for the source -> canonical and canonical
    -> model exposure transforms, etc.
    """
    def __init__(
        self,
        model_supplier_id,
        model_id,
        model_version,
        resources=None
    ):
        """
        Constructor - requires supplier ID, model ID and model version ID.
        """
        self._supplier_id = model_supplier_id
        self._model_id = model_id
        self._model_version_id = model_version
        self._key = '{}/{}/{}'.format(model_supplier_id, model_id, model_version)
        self._resources = resources if resources else {}

    def __str__(self):
        return '{}: {}'.format(self.__repr__(), self.key)

    def __repr__(self):
        return '{}: {}'.format(self.__class__, self.__dict__)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else '...')

    @property
    def key(self):
        """
        Model key - getter only. Format is

            :getter: Returns <model supplier ID>/<model ID>/<model version ID>
            :type: string
        """
        return self._key

    @property
    def supplier_id(self):
        """
        Model supplier ID property - getter only.

            :getter: Gets the model supplier ID
            :type: string
        """
        return self._supplier_id

    @property
    def model_id(self):
        """
        Model ID property - getter only.

            :getter: Gets the model ID
            :type: string
        """
        return self._model_id

    @property
    def model_version(self):
        """
        Model version ID property - getter only.

            :getter: Gets the model version ID
            :type: string
        """
        return self._model_version_id

    @property
    def resources(self):
        """
        Model resources dictionary property.

            :getter: Gets the attached resource in the model resources dict
                     using the optional resource ``key`` argument. If ``key``
                     is not given then the entire resources dict is returned.

            :setter: Sets the value of the optional resource ``key`` in the
                     resources dict to ``val``. If no ``key`` is given then
                     ``val`` is assumed to be a new resources dict and is
                     used to replace the existing dict.

            :deleter: Deletes the value of the optional resource ``key`` in
                      the resources dict. If no ``key`` is given then the
                      entire existing dict is cleared.
        """
        return self._resources

    @resources.setter
    def resources(self, val=None):
        self._resources.clear()
        self._resources.update(val)

    @resources.deleter
    def resources(self):
        self._resources.clear()


class OasisModelFactory(object):
    """
    Factory class for creating Oasis model objects.
    """
    @classmethod
    def create(
        cls,
        model_supplier_id,
        model_id,
        model_version,
        resources=None
    ):
        """
        Service method to instantiate Oasis model objects with attached
        resource dicts.
        """
        return OasisModel(
            model_supplier_id=model_supplier_id,
            model_id=model_id,
            model_version=model_version,
            resources=resources
        )



from __future__ import unicode_literals

import json
import os
import io
import string
from random import choice
from tempfile import NamedTemporaryFile
from backports.tempfile import TemporaryDirectory
from unittest import TestCase

import responses
from hypothesis import given
from hypothesis.strategies import integers
from mock import patch, Mock
from pathlib2 import Path
from requests import RequestException

from oasislmf.api_client.client import OasisAPIClient
from oasislmf.model_execution.files import TAR_FILE
from oasislmf.utils.exceptions import OasisException
from oasislmf.utils.status import STATUS_FAILURE, STATUS_PENDING, STATUS_SUCCESS


class ClientHealthCheck(TestCase):
    @given(integers(min_value=1, max_value=5))
    def test_heath_check_raise_an_exception_on_each_call___result_is_false(self, max_attempts):
        with patch('requests.get', Mock(side_effect=RequestException())) as get_mock:
            client = OasisAPIClient('http://localhost:8001')

            result = client.health_check(max_attempts, retry_delay=0)

            self.assertFalse(result)
            self.assertEqual(max_attempts, get_mock.call_count)
            call_urls = [args[0] for args in get_mock.call_args_list]
            self.assertEqual(
                [('http://localhost:8001/healthcheck',) for i in range(max_attempts)],
                call_urls,
            )

    @given(integers(min_value=1, max_value=5))
    def test_heath_check_returns_non_200_on_each_call___result_is_false(self, max_attempts):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'http://localhost:8001/healthcheck', status=404)

            client = OasisAPIClient('http://localhost:8001')

            result = client.health_check(max_attempts, retry_delay=0)

            self.assertFalse(result)
            self.assertEqual(max_attempts, len(rsps.calls))
            self.assertEqual(
                ['http://localhost:8001/healthcheck' for i in range(max_attempts)],
                [call.request.url for call in rsps.calls],
            )

    def test_heath_check_returns_200___result_is_true(self):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'http://localhost:8001/healthcheck', status=200)

            client = OasisAPIClient('http://localhost:8001')

            result = client.health_check()

            self.assertTrue(result)
            self.assertEqual(1, len(rsps.calls))
            self.assertEqual('http://localhost:8001/healthcheck', rsps.calls[0].request.url)


class RunAnalysis(TestCase):
    def test_request_response_is_not_ok___exception_is_raised(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, 'http://localhost:8001/analysis/foo', status=400)

            with self.assertRaises(OasisException):
                client.run_analysis({'analysis': 'data'}, 'foo')

            self.assertEqual(1, len(rsps.calls))
            self.assertEqual('http://localhost:8001/analysis/foo', rsps.calls[0].request.url)
            self.assertEqual({'analysis': 'data'}, json.loads(rsps.calls[0].request.body.decode()))

    def test_request_response_is_ok___response_location_is_returned(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, 'http://localhost:8001/analysis/foo', status=200, body=b'{"location": "output-location"}')

            self.assertEqual('output-location', client.run_analysis({'analysis': 'data'}, 'foo'))


class RunAnalysisAndPoll(TestCase):
    analysis_status_location = 'analysis_status_location'
    analysis_output_location = 'analysis_output_location'
    input_location = 'input_location'
    output_location = 'input_location'
    analysis_settings = {'analysis': 'settings'}

    def get_mocked_client(self):
        client = OasisAPIClient('http://localhost:8001')
        client.run_analysis = Mock(return_value=self.analysis_status_location)
        client.get_analysis_status = Mock(return_value=(STATUS_SUCCESS, self.analysis_output_location))
        client.download_outputs = Mock()
        client.delete_exposure = Mock()
        client.delete_outputs = Mock()
        return client

    def test_data_is_ready_at_first_status_call___data_is_downloaded_and_cleaned_correctly_never_sleeping(self):
        with patch('oasislmf.api_client.client.time.sleep') as sleep_mock:
            client = self.get_mocked_client()

            client.run_analysis_and_poll(self.analysis_settings, self.input_location, self.output_location)

            client.run_analysis.assert_called_once_with(self.analysis_settings, self.input_location)
            client.get_analysis_status.assert_called_once_with(self.analysis_status_location)
            sleep_mock.assert_not_called()
            client.download_outputs.assert_called_once_with(self.analysis_output_location, os.path.join(self.output_location, self.analysis_output_location + '.tar.gz'))
            client.delete_exposure.assert_called_once_with(self.input_location)
            client.delete_outputs.assert_called_once_with(self.analysis_output_location)

    @given(integers(min_value=1, max_value=10), integers(min_value=1, max_value=10))
    def test_data_is_ready_after_multiple_status_calls___data_is_downloaded_and_ceaned_correctly_sleeping_after_every_pending_call(self, pending_calls, poll_time):
        with patch('oasislmf.api_client.client.time.sleep') as sleep_mock:
            client = self.get_mocked_client()
            client.get_analysis_status = Mock(side_effect=[(STATUS_PENDING, '')] * pending_calls + [(STATUS_SUCCESS, self.analysis_output_location)])

            client.run_analysis_and_poll(self.analysis_settings, self.input_location, self.output_location, analysis_poll_interval=poll_time)

            client.run_analysis.assert_called_once_with(self.analysis_settings, self.input_location)

            self.assertEqual(pending_calls + 1, client.get_analysis_status.call_count)
            self.assertEqual([(self.analysis_status_location, )] * (pending_calls + 1), [args[0] for args in client.get_analysis_status.call_args_list])

            self.assertEqual(pending_calls, sleep_mock.call_count)
            self.assertEqual([(poll_time, )] * pending_calls, [args[0] for args in sleep_mock.call_args_list])

            client.download_outputs.assert_called_once_with(self.analysis_output_location, os.path.join(self.output_location, self.analysis_output_location + '.tar.gz'))
            client.delete_exposure.assert_called_once_with(self.input_location)
            client.delete_outputs.assert_called_once_with(self.analysis_output_location)


class GetAnalysisStatus(TestCase):
    def test_request_is_not_ok___exception_is_raised(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'http://localhost:8001/analysis_status/foo', status=400)

            with self.assertRaises(OasisException):
                client.get_analysis_status('foo')

    def test_request_status_is_failure___exception_is_raised(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                'http://localhost:8001/analysis_status/foo',
                status=200,
                body=json.dumps({'status': STATUS_FAILURE, 'message': 'oops'}).encode()
            )

            with self.assertRaises(OasisException):
                client.get_analysis_status('foo')

    def test_request_status_is_pending___result_status_is_pending_location_is_empty(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                'http://localhost:8001/analysis_status/foo',
                status=200,
                body=json.dumps({'status': STATUS_PENDING}).encode()
            )

            status, outputs_location = client.get_analysis_status('foo')

            self.assertEqual(status, STATUS_PENDING)
            self.assertEqual(outputs_location, '')

    def test_request_status_is_success___result_status_is_success_location_is_from_response_body(self):
        client = OasisAPIClient('http://localhost:8001')

        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                'http://localhost:8001/analysis_status/foo',
                status=200,
                body=json.dumps({'status': STATUS_SUCCESS, 'outputs_location': 'outputs-location'}).encode()
            )

            status, outputs_location = client.get_analysis_status('foo')

            self.assertEqual(status, STATUS_SUCCESS)
            self.assertEqual(outputs_location, 'outputs-location')


class DeleteResource(TestCase):
    def test_response_is_not_ok___warning_is_logged(self):
        client = OasisAPIClient('http://localhost:8001', Mock())

        with responses.RequestsMock() as rsps:
            rsps.add(responses.DELETE, 'http://localhost:8001/foo', status=400)

            client.delete_resource('foo')

            client._logger.warning.assert_called_with("DELETE http://localhost:8001/foo failed: 400")

    def test_response_is_ok___info_is_logged(self):
        client = OasisAPIClient('http://localhost:8001', Mock())

        with responses.RequestsMock() as rsps:
            rsps.add(responses.DELETE, 'http://localhost:8001/foo', status=200)

            client.delete_resource('foo')

            client._logger.info.assert_called_with('Deleted http://localhost:8001/foo')


class DeleteOutputs(TestCase):
    def test_delete_resource_is_called_with_the_correct_parameters(self):
        client = OasisAPIClient('http://localhost:8001')
        client.delete_resource = Mock()

        client.delete_outputs('foo')

        client.delete_resource.assert_called_once_with('/outputs/foo')


class DeleteExposures(TestCase):
    def test_delete_resource_is_called_with_the_correct_parameters(self):
        client = OasisAPIClient('http://localhost:8001')
        client.delete_resource = Mock()

        client.delete_exposure('foo')

        client.delete_resource.assert_called_once_with('/exposure/foo')


class DownloadResource(TestCase):
    def test_local_file_already_exists___exception_is_raised_and_file_is_unchanged(self):
        client = OasisAPIClient('http://localhost:8001')

        with NamedTemporaryFile('w+') as f:
            f.write('foobarboo')
            f.flush()

            with self.assertRaises(OasisException):
                client.download_resource('foo', f.name)

            f.seek(0)
            self.assertEqual('foobarboo', f.read())

    def test_response_is_not_ok___exception_is_raised_and_file_is_not_created(self):
        client = OasisAPIClient('http://localhost:8001')

        with NamedTemporaryFile('w') as f:
            local_filename = f.name

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'http://localhost:8001/foo', status=400)

            with self.assertRaises(OasisException):
                client.download_resource('foo', local_filename)

            self.assertFalse(os.path.exists(local_filename))

    def test_response_is_ok___file_is_created_with_correct_content(self):
        client = OasisAPIClient('http://localhost:8001')
        client.DOWNLOAD_CHUCK_SIZE_IN_BYTES = 10

        expected_content = ''.join(choice(string.ascii_letters) for i in range(2 * client.DOWNLOAD_CHUCK_SIZE_IN_BYTES)).encode()

        with NamedTemporaryFile('w') as f:
            local_filename = f.name

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'http://localhost:8001/foo', status=200, body=expected_content)

            client.download_resource('foo', local_filename)

            self.assertTrue(os.path.exists(local_filename))

            with io.open(local_filename, 'rb') as f:
                dld_content = f.read()

            os.remove(local_filename)
            self.assertEqual(expected_content, dld_content)


class DownloadOutputs(TestCase):
    def test_download_resource_is_called_with_the_correct_parameters(self):
        client = OasisAPIClient('http://localhost:8001')
        client.download_resource = Mock()

        client.download_outputs('foo', 'local_filename')

        client.download_resource.assert_called_once_with('/outputs/foo', 'local_filename')


class DownloadExposures(TestCase):
    def test_download_resource_is_called_with_the_correct_parameters(self):
        client = OasisAPIClient('http://localhost:8001')
        client.download_resource = Mock()

        client.download_exposure('foo', 'local_filename')

        client.download_resource.assert_called_once_with('/exposure/foo', 'local_filename')


def fake_build_tar_fn(d):
    with io.open(os.path.join(d, TAR_FILE), 'wb') as f:
        f.write(''.join(choice(string.ascii_letters) for i in range(100)).encode())


class UploadInputsFromDirectory(TestCase):
    @patch('oasislmf.api_client.client.create_binary_tar_file')
    @patch('oasislmf.api_client.client.create_binary_files')
    @patch('oasislmf.api_client.client.check_conversion_tools')
    @patch('oasislmf.api_client.client.check_inputs_directory')
    def test_do_build_is_false___bin_building_functions_are_not_called(self, check_mock, check_tools_mock, create_bin_mock, create_tar_mock):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()
            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, do_build=False)

            check_mock.assert_not_called()
            check_tools_mock.assert_not_called()
            create_bin_mock.assert_not_called()
            create_tar_mock.assert_not_called()

    @patch('oasislmf.api_client.client.create_binary_tar_file')
    @patch('oasislmf.api_client.client.create_binary_files')
    @patch('oasislmf.api_client.client.check_conversion_tools')
    @patch('oasislmf.api_client.client.check_inputs_directory')
    def test_do_build_is_true_do_il_is_false___bin_building_functions_are_called_with_correct_args(self, check_mock, check_tools_mock, create_bin_mock, create_tar_mock):
        create_tar_mock.side_effect = fake_build_tar_fn

        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, do_build=True)

            check_mock.assert_called_once_with(d, do_il=False)
            check_tools_mock.assert_called_once_with(do_il=False)
            create_bin_mock.assert_called_once_with(d, d, do_il=False)
            create_tar_mock.assert_called_once_with(d)

    @patch('oasislmf.api_client.client.create_binary_tar_file')
    @patch('oasislmf.api_client.client.create_binary_files')
    @patch('oasislmf.api_client.client.check_conversion_tools')
    @patch('oasislmf.api_client.client.check_inputs_directory')
    def test_do_build_is_true_do_il_is_true___bin_building_functions_are_called_with_correct_args(self, check_mock, check_tools_mock, create_bin_mock, create_tar_mock):
        create_tar_mock.side_effect = fake_build_tar_fn

        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, do_build=True, do_il=True)

            check_mock.assert_called_once_with(d, do_il=True)
            check_tools_mock.assert_called_once_with(do_il=True)
            create_bin_mock.assert_called_once_with(d, d, do_il=True)
            create_tar_mock.assert_called_once_with(d)

    @patch('oasislmf.api_client.client.create_binary_tar_file')
    @patch('oasislmf.api_client.client.create_binary_files')
    @patch('oasislmf.api_client.client.check_conversion_tools')
    @patch('oasislmf.api_client.client.check_inputs_directory')
    def test_do_build_is_truebin_dir_is_supplied___bin_building_functions_are_called_with_correct_args(self, check_mock, check_tools_mock, create_bin_mock, create_tar_mock):
        create_tar_mock.side_effect = fake_build_tar_fn

        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            bin_dir = os.path.join(d, 'alt_bin_dir')
            os.mkdir(bin_dir)

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, bin_directory=bin_dir, do_build=True)

            check_mock.assert_called_once_with(d, do_il=False)
            check_tools_mock.assert_called_once_with(do_il=False)
            create_bin_mock.assert_called_once_with(d, bin_dir, do_il=False)
            create_tar_mock.assert_called_once_with(bin_dir)

    def test_tar_file_exists___correct_file_is_posted(self):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d)

            self.assertEqual(1, len(rsps.calls))

            request = rsps.calls[0].request
            self.assertEqual(request.url, 'http://localhost:8001/exposure')

            multipart_data = request.body.fields['file']
            self.assertEqual('inputs.tar.gz', multipart_data[0])
            self.assertEqual(os.path.join(d, TAR_FILE), multipart_data[1].name)
            self.assertEqual('text/plain', multipart_data[2])

    def test_tar_file_exists_in_specified_bin_dir___correct_file_is_posted(self):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            bin_dir = os.path.join(d, 'alt_bin_dir')
            os.mkdir(bin_dir)
            Path(os.path.join(bin_dir, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, bin_directory=bin_dir)

            self.assertEqual(1, len(rsps.calls))

            request = rsps.calls[0].request
            self.assertEqual(request.url, 'http://localhost:8001/exposure')

            multipart_data = request.body.fields['file']
            self.assertEqual('inputs.tar.gz', multipart_data[0])
            self.assertEqual(os.path.join(bin_dir, TAR_FILE), multipart_data[1].name)
            self.assertEqual('text/plain', multipart_data[2])

    def test_tar_file_exists___exposure_location_from_result_is_returned(self):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            result = client.upload_inputs_from_directory(d)

            self.assertEqual('exposure_location', result)

    def test_response_from_server_is_not_ok___oasis_error_is_raised(self):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()

            rsps.add(responses.POST, status=400, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')

            with self.assertRaises(OasisException):
                client.upload_inputs_from_directory(d)

    @patch('oasislmf.api_client.client.cleanup_bin_directory')
    def test_do_clean_is_false___clean_bin_directory_is_not_called(self, clean_mock):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d)

            clean_mock.assert_not_called()

    @patch('oasislmf.api_client.client.cleanup_bin_directory')
    def test_do_clean_is_true___clean_bin_directory_is_called_on_the_correct_directory(self, clean_mock):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            Path(os.path.join(d, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, do_clean=True)

            clean_mock.assert_called_once_with(d)

    @patch('oasislmf.api_client.client.cleanup_bin_directory')
    def test_do_clean_is_true_and_bin_dir_is_specified___clean_bin_directory_is_called_on_the_correct_directory(self, clean_mock):
        with TemporaryDirectory() as d, responses.RequestsMock() as rsps:
            bin_dir = os.path.join(d, 'alt_bin_dir')
            os.mkdir(bin_dir)
            Path(os.path.join(bin_dir, TAR_FILE)).touch()

            rsps.add(responses.POST, url='http://localhost:8001/exposure', body=json.dumps({'exposures': [{'location': 'exposure_location'}]}).encode())

            client = OasisAPIClient('http://localhost:8001')
            client.upload_inputs_from_directory(d, bin_directory=bin_dir, do_clean=True)

            clean_mock.assert_called_once_with(bin_dir)
